import asyncio
import os
from typing import Optional

from google.protobuf.json_format import Parse, ParseError

from ..interactive_tree import interactive_tree
from ..tools.cli_tools import *
from ...datatypes import datatypes
from ...core import errors
from ..tools.click_wrappers import WrapperGroup
from ..tools.interactive_actions import ask_question_with_context, contact_new, print_with_context


#####
# identity
#####
@interactive_tree.group("identity", help="select current identity, (admin) manage olvid identities", cls=WrapperGroup)
def identity_tree():
	pass


#####
# identity current
#####
@identity_tree.command("current", help="get or update identity currently used for your commands")
@click.argument("identity_id", nargs=1, type=click.INT, required=False)
async def identity_current(identity_id: int):
	# If no arguments show current identity
	if (not identity_id):
		if not ClientSingleton.get_current_identity_id():
			print_command_result("No current identity")
		else:
			identity = await ClientSingleton.get_client().identity_get()
			print_command_result("Identity currently used: {}".format(identity.details))
		return

	if not ClientSingleton.is_client_admin():
		print_error_message("Cannot change current identity when impersonating a non admin key")
		return

	# unset current identity
	if identity_id == 0:
		ClientSingleton.set_current_identity_id(0)
		print_command_result("Unselected current identity")
		return

	# change current identity index (save previous version)
	previous_identity_id = ClientSingleton.get_current_identity_id()
	ClientSingleton.set_current_identity_id(identity_id)

	try:
		# check if identity exists
		identity = await ClientSingleton.get_client().identity_get()
		# hide invitation url and show result
		identity.invitation_url = ""
		print_normal_message(identity, identity.id)
	except errors.NotFoundError as e:
		ClientSingleton.set_current_identity_id(previous_identity_id)
		raise e


#####
# identity new
#####
@identity_tree.command("new", help="create a new identity")
@click.option("-f", "--first", "first_opt", default="", help="first name")
@click.option("-l", "--last", "last_opt", default="", help="last name")
@click.option("-p", "--position", "position_opt", default="", help="position")
@click.option("-c", "--company", "company_opt", default="", help="company")
@click.option("-s", "--server", "server_url", default="", help="custom server url (optional)")
@click.argument("firstName", required=False)
@click.argument("lastName", required=False)
@click.argument("position", required=False)
@click.argument("company", required=False)
async def identity_new(first_opt: str, last_opt: str, position_opt: str, company_opt: str, firstname: str, lastname: str, position: str, company: str, server_url: str):
	if not ClientSingleton.is_client_admin():
		print_error_message("Cannot create an identity when impersonating a non admin key")
		return

	identity_details = datatypes.IdentityDetails()
	if (first_opt or firstname):
		identity_details.first_name = first_opt if first_opt else firstname
	if (last_opt or lastname):
		identity_details.last_name = last_opt if last_opt else lastname
	if (company_opt or company):
		identity_details.company = company_opt if company_opt else company
	if (position_opt or position):
		identity_details.position = position_opt if position_opt else position

	identity = await ClientSingleton.get_client().admin_identity_new(identity_details=identity_details, server_url=server_url)

	if ClientSingleton.is_script_mode_enabled():
		print(identity.id)
		return

	prompt: str = "identity creation"
	fg_color: str = "bright_blue"

	print_command_result(f"identity created with id: {identity.id}")

	try:
		# create associated client key
		client_key: datatypes.ClientKey = await ClientSingleton.get_client().admin_client_key_new(name=f"identity-new-key-{identity.id}", identity_id=identity.id)
		print_with_context(f"Here is your client key to connect to daemon with this identity:\n{click.style(client_key.key, underline=True)}", prompt=prompt, fg_color=fg_color)
		await asyncio.sleep(0.5)

		# ask to add this new identity as a contact
		if not await ask_question_with_context("Do you want to add this identity to your contacts ?", prompt=prompt, fg_color=fg_color):
			ClientSingleton.set_current_identity_id(identity.id)
			click.secho(f"Now using identity: {ClientSingleton.get_current_identity_id()}", fg="green")
			return

		# add this new identity as a contact
		discussion: Optional[datatypes.Discussion] = await contact_new(identity_id=identity.id, prompt=prompt, fg_color=fg_color)

		ClientSingleton.set_current_identity_id(identity.id)
		click.secho(f"Now using identity: {ClientSingleton.get_current_identity_id()}", fg="green")

		if discussion:
			print(f"You can now send messages to {discussion.title} in discussion {discussion.id}")
		else:
			print("Invitation process finished")
	except click.exceptions.Abort:
		pass


#####
# identity get
#####
# noinspection PyProtectedMember
@identity_tree.command("get", help="list identities on this daemon")
@click.option("-a", "--all", "get_all", is_flag=True)
@click.option("-l", "--link", "show_invitation_link", is_flag=True)
@click.option("-i", "--identifier", "show_identity_identifier", is_flag=True)
@click.option("-f", "--fields", "fields", type=str)
@click.argument("identity_ids", nargs=-1, type=click.INT)
@click.option("--filter", "filter_", type=str)
async def identity_get(get_all: bool, show_invitation_link: bool, show_identity_identifier: bool, identity_ids: tuple[int], fields: str, filter_: str = ""):
	# normal case
	if ClientSingleton.is_client_admin():
		# build filter
		identity_filter: datatypes.IdentityFilter = datatypes.IdentityFilter()
		if filter_:
			try:
				parsed_message = Parse(filter_, datatypes.IdentityFilter()._to_native(identity_filter))
				identity_filter = datatypes.IdentityFilter._from_native(parsed_message)
			except ParseError as e:
				print_error_message(f"Cannot parse filter: {e}")
				return

		if get_all or not identity_ids:
			identities: list[datatypes.Identity] = [i async for i in ClientSingleton.get_client().admin_identity_list(filter=identity_filter)]
		else:
			identities: list[datatypes.Identity] = [await ClientSingleton.get_client().admin_identity_admin_get(identity_id) for identity_id in identity_ids]
	# non-admin client case
	else:
		if identity_ids:
			print_error_message("Cannot list identities when impersonating a non admin key")
			return
		identities: list[datatypes.Identity] = [await ClientSingleton.get_client().identity_get()]

	# hide deprecated invitation url field
	for identity in identities:
		identity.invitation_url = ""
		if show_invitation_link:
			original_current_identity_id = ClientSingleton.get_current_identity_id()
			ClientSingleton.set_current_identity_id(identity.id)
			print(f"{identity.id}: {await ClientSingleton.get_client().identity_get_invitation_link()}")
			ClientSingleton.set_current_identity_id(original_current_identity_id)
		elif show_identity_identifier:
			original_current_identity_id = ClientSingleton.get_current_identity_id()
			ClientSingleton.set_current_identity_id(identity.id)
			print(f"{identity.id}: {await ClientSingleton.get_client().identity_get_bytes_identifier()}")
			ClientSingleton.set_current_identity_id(original_current_identity_id)
		else:
			filter_fields_and_print_normal_message(identity, fields)


#####
# identity rm
#####
@identity_tree.command("rm", help="delete identities")
@click.option("-a", "--all", "delete_all", is_flag=True)
@click.option("-e", "--everywhere", is_flag=True, help="delete identity on all your devices and notify contacts")
@click.argument("identity_ids", nargs=-1, type=click.INT)
async def identity_delete(delete_all: bool, identity_ids: tuple[int], everywhere: bool):
	if not ClientSingleton.is_client_admin():
		print_error_message("Cannot delete an identity when impersonating a non admin key")
		return

	identity_ids = list(identity_ids)  # convert tuple to list
	if (not len(identity_ids) and not delete_all):
		raise click.exceptions.BadArgumentUsage("Specify identity id")

	if delete_all:
		identity_ids = [i.id async for i in ClientSingleton.get_client().admin_identity_list()]
		if not identity_ids:
			raise click.exceptions.BadArgumentUsage("No identity to delete")

	for identity_id in identity_ids:
		await ClientSingleton.get_client().admin_identity_delete(identity_id, delete_everywhere=everywhere)
		print_command_result(f"Identity deleted: {identity_id}", identity_id)


#####
# identity update
#####
@identity_tree.command("update", help="update your current identity details (name, position, ...)")
@click.option("-f", "--first", "first_opt", default="")
@click.option("-l", "--last", "last_opt", default="")
@click.option("-p", "--position", "position_opt", default="")
@click.option("-c", "--company", "company_opt", default="")
@click.argument("firstName", required=False)
@click.argument("lastName", required=False)
@click.argument("position", required=False)
@click.argument("company", required=False)
async def identity_update_details(first_opt: str, last_opt: str, position_opt: str, company_opt: str, firstname: str, lastname: str, position: str, company: str):
	identity_details = datatypes.IdentityDetails()
	if (first_opt or firstname):
		identity_details.first_name = first_opt if first_opt else firstname
	if (last_opt or lastname):
		identity_details.last_name = last_opt if last_opt else lastname
	if (company_opt or company):
		identity_details.company = company_opt if company_opt else company
	if (position_opt or position):
		identity_details.position = position_opt if position_opt else position

	if not identity_details.first_name and not identity_details.last_name:
		raise click.exceptions.BadArgumentUsage("Specify at least first or last name")

	await ClientSingleton.get_client().identity_update_details(identity_details)
	print_command_result("Identity updated")


#####
# identity key
#####
@identity_tree.group("key", help="manage current identity Olvid api key", cls=WrapperGroup)
def identity_key_tree():
	pass

#####
# identity key get
#####
@identity_key_tree.command("get", help="get api key for current identity")
async def identity_apikey_get():
	result = await ClientSingleton.get_client().identity_get_api_key_status()
	if result:
		print_command_result(f"Current api key: {result}")
	else:
		print_command_result(f"Api key not set")

#####
# identity key set
#####
@identity_key_tree.command("set", help="set api key for current identity")
@click.argument("apiKey", nargs=1, required=True)
@click.option("-c", "--configuration", is_flag=True, required=False, help="Use a configuration link instead of a raw api key")
async def identity_apikey_set(configuration: bool, apikey: str):
	if configuration:
		result = await ClientSingleton.get_client().identity_set_configuration_link(configuration_link=apikey)
	else:
		result = await ClientSingleton.get_client().identity_set_api_key(api_key=apikey)
	print_command_result(f"Api key set up, new permissions: {result}")


#####
# identity photo
#####
@identity_tree.group("photo", help="manage current identity photo", cls=WrapperGroup)
def identity_photo_tree():
	pass


#####
# identity photo set
#####
@identity_photo_tree.command("set", help="set current identity photo")
@click.argument("photo_path", required=True, type=click.STRING)
async def identity_photo_set(photo_path):
	try:
		await ClientSingleton.get_client().identity_set_photo_file(photo_path)
		print_command_result("Identity photo set")
	except IOError as e:
		raise click.exceptions.BadArgumentUsage(str(e))


#####
# identity photo save
#####
@identity_photo_tree.command("save", help="Save identity photo to local files. Specify identity_ids to use or it uses current identity id by default")
@click.argument("identity_ids", required=False, nargs=-1, type=click.INT)
@click.option("-a", "--all", "save_all", is_flag=True, help="save all identity photos")
@click.option("-p", "--path", "path", help="directory to store downloaded photo (default: ./photos)", nargs=1, type=click.STRING, required=False)
@click.option("-f", "--filename", "filename", help="specify file name to use (ignored if saving more than one image)", nargs=1, type=click.STRING, required=False)
async def identity_photo_set(path: str, filename: str, save_all: bool, identity_ids: list[int]):
	# if identity_id is not specified and not --all use current identity
	if save_all:
		if not ClientSingleton.is_client_admin():
			print_error_message("Cannot save every identity photo identity when impersonating a non admin key")
			return
		identity_ids = [i.id async for i in ClientSingleton.get_client().admin_identity_list()]
	if not save_all and not identity_ids:
		# check current identity is properly set
		current_identity = await ClientSingleton.get_client().identity_get()
		identity_ids = [current_identity.id]

	# use default save directory if necessary
	if not path:
		path = "./photos"
	# create save directory
	os.makedirs(path, exist_ok=True)

	# save every requested photo
	for identity_id in identity_ids:
		photo_bytes = await ClientSingleton.get_client().admin_identity_admin_download_photo(identity_id=identity_id)

		# specified filename flag
		if len(identity_ids) == 1 and filename:
			filepath: str = os.path.join(path, filename)
		# default filename
		else:
			filepath: str = os.path.join(path, f"identity_{identity_id}.jpeg")
		with open(filepath, "wb") as photo:
			photo.write(photo_bytes)
		print_normal_message(f"Photo saved in: {filepath}", filepath)


#####
# identity photo unset
#####
@identity_photo_tree.command("unset", help="remove current identity photo")
async def identity_photo_unset():
	await ClientSingleton.get_client().identity_remove_photo()
	print_command_result("Identity photo unset")


#####
# identity kc
#####
@identity_tree.group("kc", help="manage keycloak server for identities", cls=WrapperGroup)
def identity_kc_tree():
	pass


#####
# identity kc new
#####
@identity_kc_tree.command("new", help="create a new identity using a keycloak bot configuration link")
@click.argument("configuration_link", required=True, type=click.STRING)
async def identity_kc_new(configuration_link):
	if not ClientSingleton.is_client_admin():
		print_error_message("Cannot create a new identity when impersonating a non admin key")
		return

	identity = await ClientSingleton.get_client().admin_identity_keycloak_new(configuration_link=configuration_link)

	if ClientSingleton.is_script_mode_enabled():
		print(identity.id)
		return

	prompt: str = "identity creation"
	fg_color: str = "bright_blue"

	print_command_result(f"identity created with id: {identity.id}")

	try:
		# create associated client key
		client_key: datatypes.ClientKey = await ClientSingleton.get_client().admin_client_key_new(name=f"identity-new-key-{identity.id}", identity_id=identity.id)
		print_with_context(f"Here is your client key to connect to daemon with this identity:\n{click.style(client_key.key, underline=True)}", prompt=prompt, fg_color=fg_color)
		await asyncio.sleep(0.5)

		ClientSingleton.set_current_identity_id(identity.id)
		click.secho(f"Now using identity: {ClientSingleton.get_current_identity_id()}", fg="green")
	except click.exceptions.Abort:
		pass



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: olvid/daemon/datatypes/v1/identity.proto
# Protobuf Python Version: 6.33.2
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    6,
    33,
    2,
    '',
    'olvid/daemon/datatypes/v1/identity.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n(olvid/daemon/datatypes/v1/identity.proto\x12\x19olvid.daemon.datatypes.v1\"\xcd\x01\n\x0fIdentityDetails\x12\"\n\nfirst_name\x18\x01 \x01(\tH\x00R\tfirstName\x88\x01\x01\x12 \n\tlast_name\x18\x02 \x01(\tH\x01R\x08lastName\x88\x01\x01\x12\x1d\n\x07\x63ompany\x18\x03 \x01(\tH\x02R\x07\x63ompany\x88\x01\x01\x12\x1f\n\x08position\x18\x04 \x01(\tH\x03R\x08position\x88\x01\x01\x42\r\n\x0b_first_nameB\x0c\n\n_last_nameB\n\n\x08_companyB\x0b\n\t_position\"\x98\x04\n\x08Identity\x12\x0e\n\x02id\x18\x01 \x01(\x04R\x02id\x12!\n\x0c\x64isplay_name\x18\x02 \x01(\tR\x0b\x64isplayName\x12\x44\n\x07\x64\x65tails\x18\x03 \x01(\x0b\x32*.olvid.daemon.datatypes.v1.IdentityDetailsR\x07\x64\x65tails\x12)\n\x0einvitation_url\x18\x04 \x01(\tB\x02\x18\x01R\rinvitationUrl\x12)\n\x10keycloak_managed\x18\x05 \x01(\x08R\x0fkeycloakManaged\x12\x1e\n\x0bhas_a_photo\x18\x06 \x01(\x08R\thasAPhoto\x12\x43\n\x07\x61pi_key\x18\x07 \x01(\x0b\x32*.olvid.daemon.datatypes.v1.Identity.ApiKeyR\x06\x61piKey\x1a\xd7\x01\n\x06\x41piKey\x12U\n\npermission\x18\x01 \x01(\x0b\x32\x35.olvid.daemon.datatypes.v1.Identity.ApiKey.PermissionR\npermission\x12\x31\n\x14\x65xpiration_timestamp\x18\x02 \x01(\x04R\x13\x65xpirationTimestamp\x1a\x43\n\nPermission\x12\x12\n\x04\x63\x61ll\x18\x01 \x01(\x08R\x04\x63\x61ll\x12!\n\x0cmulti_device\x18\x02 \x01(\x08R\x0bmultiDevice\"\xb3\x05\n\x0eIdentityFilter\x12S\n\x08keycloak\x18\x01 \x01(\x0e\x32\x32.olvid.daemon.datatypes.v1.IdentityFilter.KeycloakH\x00R\x08keycloak\x88\x01\x01\x12J\n\x05photo\x18\x02 \x01(\x0e\x32/.olvid.daemon.datatypes.v1.IdentityFilter.PhotoH\x01R\x05photo\x88\x01\x01\x12N\n\x07\x61pi_key\x18\x03 \x01(\x0e\x32\x30.olvid.daemon.datatypes.v1.IdentityFilter.ApiKeyH\x02R\x06\x61piKey\x88\x01\x01\x12\x33\n\x13\x64isplay_name_search\x18\x04 \x01(\tH\x03R\x11\x64isplayNameSearch\x88\x01\x01\x12V\n\x0e\x64\x65tails_search\x18\x05 \x01(\x0b\x32*.olvid.daemon.datatypes.v1.IdentityDetailsH\x04R\rdetailsSearch\x88\x01\x01\"J\n\x08Keycloak\x12\x18\n\x14KEYCLOAK_UNSPECIFIED\x10\x00\x12\x0f\n\x0bKEYCLOAK_IS\x10\x02\x12\x13\n\x0fKEYCLOAK_IS_NOT\x10\x01\"@\n\x05Photo\x12\x15\n\x11PHOTO_UNSPECIFIED\x10\x00\x12\r\n\tPHOTO_HAS\x10\x02\x12\x11\n\rPHOTO_HAS_NOT\x10\x01\"G\n\x06\x41piKey\x12\x17\n\x13\x41PI_KEY_UNSPECIFIED\x10\x00\x12\x0f\n\x0b\x41PI_KEY_HAS\x10\x02\x12\x13\n\x0f\x41PI_KEY_HAS_NOT\x10\x01\x42\x0b\n\t_keycloakB\x08\n\x06_photoB\n\n\x08_api_keyB\x16\n\x14_display_name_searchB\x11\n\x0f_details_searchB\xc6\x01\n\x1d\x63om.olvid.daemon.datatypes.v1B\rIdentityProtoP\x01Z\x0folvid.io/daemon\xa2\x02\x03ODD\xaa\x02\x19Olvid.Daemon.Datatypes.V1\xca\x02\x19Olvid\\Daemon\\Datatypes\\V1\xe2\x02%Olvid\\Daemon\\Datatypes\\V1\\GPBMetadata\xea\x02\x1cOlvid::Daemon::Datatypes::V1b\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'olvid.daemon.datatypes.v1.identity_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\035com.olvid.daemon.datatypes.v1B\rIdentityProtoP\001Z\017olvid.io/daemon\242\002\003ODD\252\002\031Olvid.Daemon.Datatypes.V1\312\002\031Olvid\\Daemon\\Datatypes\\V1\342\002%Olvid\\Daemon\\Datatypes\\V1\\GPBMetadata\352\002\034Olvid::Daemon::Datatypes::V1'
  _globals['_IDENTITY'].fields_by_name['invitation_url']._loaded_options = None
  _globals['_IDENTITY'].fields_by_name['invitation_url']._serialized_options = b'\030\001'
  _globals['_IDENTITYDETAILS']._serialized_start=72
  _globals['_IDENTITYDETAILS']._serialized_end=277
  _globals['_IDENTITY']._serialized_start=280
  _globals['_IDENTITY']._serialized_end=816
  _globals['_IDENTITY_APIKEY']._serialized_start=601
  _globals['_IDENTITY_APIKEY']._serialized_end=816
  _globals['_IDENTITY_APIKEY_PERMISSION']._serialized_start=749
  _globals['_IDENTITY_APIKEY_PERMISSION']._serialized_end=816
  _globals['_IDENTITYFILTER']._serialized_start=819
  _globals['_IDENTITYFILTER']._serialized_end=1510
  _globals['_IDENTITYFILTER_KEYCLOAK']._serialized_start=1219
  _globals['_IDENTITYFILTER_KEYCLOAK']._serialized_end=1293
  _globals['_IDENTITYFILTER_PHOTO']._serialized_start=1295
  _globals['_IDENTITYFILTER_PHOTO']._serialized_end=1359
  _globals['_IDENTITYFILTER_APIKEY']._serialized_start=1361
  _globals['_IDENTITYFILTER_APIKEY']._serialized_end=1432
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: olvid/daemon/command/v1/invitation_commands.proto
# Protobuf Python Version: 6.33.2
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    6,
    33,
    2,
    '',
    'olvid/daemon/command/v1/invitation_commands.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from .....olvid.daemon.datatypes.v1 import invitation_pb2 as olvid_dot_daemon_dot_datatypes_dot_v1_dot_invitation__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n1olvid/daemon/command/v1/invitation_commands.proto\x12\x17olvid.daemon.command.v1\x1a*olvid/daemon/datatypes/v1/invitation.proto\"l\n\x15InvitationListRequest\x12H\n\x06\x66ilter\x18\x01 \x01(\x0b\x32+.olvid.daemon.datatypes.v1.InvitationFilterH\x00R\x06\x66ilter\x88\x01\x01\x42\t\n\x07_filter\"a\n\x16InvitationListResponse\x12G\n\x0binvitations\x18\x01 \x03(\x0b\x32%.olvid.daemon.datatypes.v1.InvitationR\x0binvitations\";\n\x14InvitationGetRequest\x12#\n\rinvitation_id\x18\x01 \x01(\x04R\x0cinvitationId\"^\n\x15InvitationGetResponse\x12\x45\n\ninvitation\x18\x01 \x01(\x0b\x32%.olvid.daemon.datatypes.v1.InvitationR\ninvitation\"=\n\x14InvitationNewRequest\x12%\n\x0einvitation_url\x18\x01 \x01(\tR\rinvitationUrl\"^\n\x15InvitationNewResponse\x12\x45\n\ninvitation\x18\x01 \x01(\x0b\x32%.olvid.daemon.datatypes.v1.InvitationR\ninvitation\">\n\x17InvitationAcceptRequest\x12#\n\rinvitation_id\x18\x01 \x01(\x04R\x0cinvitationId\"\x1a\n\x18InvitationAcceptResponse\"?\n\x18InvitationDeclineRequest\x12#\n\rinvitation_id\x18\x01 \x01(\x04R\x0cinvitationId\"\x1b\n\x19InvitationDeclineResponse\"M\n\x14InvitationSasRequest\x12#\n\rinvitation_id\x18\x01 \x01(\x04R\x0cinvitationId\x12\x10\n\x03sas\x18\x02 \x01(\tR\x03sas\"\x17\n\x15InvitationSasResponse\">\n\x17InvitationDeleteRequest\x12#\n\rinvitation_id\x18\x01 \x01(\x04R\x0cinvitationId\"\x1a\n\x18InvitationDeleteResponseB\xc6\x01\n\x1b\x63om.olvid.daemon.command.v1B\x17InvitationCommandsProtoP\x01Z\x0folvid.io/daemon\xa2\x02\x03ODC\xaa\x02\x17Olvid.Daemon.Command.V1\xca\x02\x17Olvid\\Daemon\\Command\\V1\xe2\x02#Olvid\\Daemon\\Command\\V1\\GPBMetadata\xea\x02\x1aOlvid::Daemon::Command::V1b\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'olvid.daemon.command.v1.invitation_commands_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\033com.olvid.daemon.command.v1B\027InvitationCommandsProtoP\001Z\017olvid.io/daemon\242\002\003ODC\252\002\027Olvid.Daemon.Command.V1\312\002\027Olvid\\Daemon\\Command\\V1\342\002#Olvid\\Daemon\\Command\\V1\\GPBMetadata\352\002\032Olvid::Daemon::Command::V1'
  _globals['_INVITATIONLISTREQUEST']._serialized_start=122
  _globals['_INVITATIONLISTREQUEST']._serialized_end=230
  _globals['_INVITATIONLISTRESPONSE']._serialized_start=232
  _globals['_INVITATIONLISTRESPONSE']._serialized_end=329
  _globals['_INVITATIONGETREQUEST']._serialized_start=331
  _globals['_INVITATIONGETREQUEST']._serialized_end=390
  _globals['_INVITATIONGETRESPONSE']._serialized_start=392
  _globals['_INVITATIONGETRESPONSE']._serialized_end=486
  _globals['_INVITATIONNEWREQUEST']._serialized_start=488
  _globals['_INVITATIONNEWREQUEST']._serialized_end=549
  _globals['_INVITATIONNEWRESPONSE']._serialized_start=551
  _globals['_INVITATIONNEWRESPONSE']._serialized_end=645
  _globals['_INVITATIONACCEPTREQUEST']._serialized_start=647
  _globals['_INVITATIONACCEPTREQUEST']._serialized_end=709
  _globals['_INVITATIONACCEPTRESPONSE']._serialized_start=711
  _globals['_INVITATIONACCEPTRESPONSE']._serialized_end=737
  _globals['_INVITATIONDECLINEREQUEST']._serialized_start=739
  _globals['_INVITATIONDECLINEREQUEST']._serialized_end=802
  _globals['_INVITATIONDECLINERESPONSE']._serialized_start=804
  _globals['_INVITATIONDECLINERESPONSE']._serialized_end=831
  _globals['_INVITATIONSASREQUEST']._serialized_start=833
  _globals['_INVITATIONSASREQUEST']._serialized_end=910
  _globals['_INVITATIONSASRESPONSE']._serialized_start=912
  _globals['_INVITATIONSASRESPONSE']._serialized_end=935
  _globals['_INVITATIONDELETEREQUEST']._serialized_start=937
  _globals['_INVITATIONDELETEREQUEST']._serialized_end=999
  _globals['_INVITATIONDELETERESPONSE']._serialized_start=1001
  _globals['_INVITATIONDELETERESPONSE']._serialized_end=1027
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: olvid/daemon/admin/v1/backup_admin.proto
# Protobuf Python Version: 6.33.2
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    6,
    33,
    2,
    '',
    'olvid/daemon/admin/v1/backup_admin.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from .....olvid.daemon.datatypes.v1 import backup_pb2 as olvid_dot_daemon_dot_datatypes_dot_v1_dot_backup__pb2
from .....olvid.daemon.datatypes.v1 import client_key_pb2 as olvid_dot_daemon_dot_datatypes_dot_v1_dot_client__key__pb2
from .....olvid.daemon.datatypes.v1 import identity_pb2 as olvid_dot_daemon_dot_datatypes_dot_v1_dot_identity__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n(olvid/daemon/admin/v1/backup_admin.proto\x12\x15olvid.daemon.admin.v1\x1a&olvid/daemon/datatypes/v1/backup.proto\x1a*olvid/daemon/datatypes/v1/client_key.proto\x1a(olvid/daemon/datatypes/v1/identity.proto\"\x15\n\x13\x42\x61\x63kupKeyGetRequest\"5\n\x14\x42\x61\x63kupKeyGetResponse\x12\x1d\n\nbackup_key\x18\x01 \x01(\tR\tbackupKey\"\x17\n\x15\x42\x61\x63kupKeyRenewRequest\"7\n\x16\x42\x61\x63kupKeyRenewResponse\x12\x1d\n\nbackup_key\x18\x01 \x01(\tR\tbackupKey\"1\n\x10\x42\x61\x63kupGetRequest\x12\x1d\n\nbackup_key\x18\x01 \x01(\tR\tbackupKey\"N\n\x11\x42\x61\x63kupGetResponse\x12\x39\n\x06\x62\x61\x63kup\x18\x01 \x01(\x0b\x32!.olvid.daemon.datatypes.v1.BackupR\x06\x62\x61\x63kup\"\x12\n\x10\x42\x61\x63kupNowRequest\"\x13\n\x11\x42\x61\x63kupNowResponse\"|\n\x1a\x42\x61\x63kupRestoreDaemonRequest\x12\x1d\n\nbackup_key\x18\x01 \x01(\tR\tbackupKey\x12+\n\x0fnew_device_name\x18\x02 \x01(\tH\x00R\rnewDeviceName\x88\x01\x01\x42\x12\n\x10_new_device_name\"\xae\x02\n\x1b\x42\x61\x63kupRestoreDaemonResponse\x12T\n\x13restored_identities\x18\x01 \x03(\x0b\x32#.olvid.daemon.datatypes.v1.IdentityR\x12restoredIdentities\x12\x61\n\x1arestored_admin_client_keys\x18\x02 \x03(\x0b\x32$.olvid.daemon.datatypes.v1.ClientKeyR\x17restoredAdminClientKeys\x12V\n\x14restored_client_keys\x18\x03 \x03(\x0b\x32$.olvid.daemon.datatypes.v1.ClientKeyR\x12restoredClientKeys\"@\n\x1f\x42\x61\x63kupRestoreAdminBackupRequest\x12\x1d\n\nbackup_key\x18\x01 \x01(\tR\tbackupKey\"\x85\x01\n BackupRestoreAdminBackupResponse\x12\x61\n\x1arestored_admin_client_keys\x18\x01 \x03(\x0b\x32$.olvid.daemon.datatypes.v1.ClientKeyR\x17restoredAdminClientKeys\"\x95\x01\n#BackupRestoreProfileSnapshotRequest\x12\x1d\n\nbackup_key\x18\x01 \x01(\tR\tbackupKey\x12\x0e\n\x02id\x18\x02 \x01(\tR\x02id\x12+\n\x0fnew_device_name\x18\x03 \x01(\tH\x00R\rnewDeviceName\x88\x01\x01\x42\x12\n\x10_new_device_name\"\xd0\x01\n$BackupRestoreProfileSnapshotResponse\x12P\n\x11restored_identity\x18\x01 \x01(\x0b\x32#.olvid.daemon.datatypes.v1.IdentityR\x10restoredIdentity\x12V\n\x14restored_client_keys\x18\x02 \x03(\x0b\x32$.olvid.daemon.datatypes.v1.ClientKeyR\x12restoredClientKeysB\xb5\x01\n\x19\x63om.olvid.daemon.admin.v1B\x10\x42\x61\x63kupAdminProtoP\x01Z\x0folvid.io/daemon\xa2\x02\x03ODA\xaa\x02\x15Olvid.Daemon.Admin.V1\xca\x02\x15Olvid\\Daemon\\Admin\\V1\xe2\x02!Olvid\\Daemon\\Admin\\V1\\GPBMetadata\xea\x02\x18Olvid::Daemon::Admin::V1b\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'olvid.daemon.admin.v1.backup_admin_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\031com.olvid.daemon.admin.v1B\020BackupAdminProtoP\001Z\017olvid.io/daemon\242\002\003ODA\252\002\025Olvid.Daemon.Admin.V1\312\002\025Olvid\\Daemon\\Admin\\V1\342\002!Olvid\\Daemon\\Admin\\V1\\GPBMetadata\352\002\030Olvid::Daemon::Admin::V1'
  _globals['_BACKUPKEYGETREQUEST']._serialized_start=193
  _globals['_BACKUPKEYGETREQUEST']._serialized_end=214
  _globals['_BACKUPKEYGETRESPONSE']._serialized_start=216
  _globals['_BACKUPKEYGETRESPONSE']._serialized_end=269
  _globals['_BACKUPKEYRENEWREQUEST']._serialized_start=271
  _globals['_BACKUPKEYRENEWREQUEST']._serialized_end=294
  _globals['_BACKUPKEYRENEWRESPONSE']._serialized_start=296
  _globals['_BACKUPKEYRENEWRESPONSE']._serialized_end=351
  _globals['_BACKUPGETREQUEST']._serialized_start=353
  _globals['_BACKUPGETREQUEST']._serialized_end=402
  _globals['_BACKUPGETRESPONSE']._serialized_start=404
  _globals['_BACKUPGETRESPONSE']._serialized_end=482
  _globals['_BACKUPNOWREQUEST']._serialized_start=484
  _globals['_BACKUPNOWREQUEST']._serialized_end=502
  _globals['_BACKUPNOWRESPONSE']._serialized_start=504
  _globals['_BACKUPNOWRESPONSE']._serialized_end=523
  _globals['_BACKUPRESTOREDAEMONREQUEST']._serialized_start=525
  _globals['_BACKUPRESTOREDAEMONREQUEST']._serialized_end=649
  _globals['_BACKUPRESTOREDAEMONRESPONSE']._serialized_start=652
  _globals['_BACKUPRESTOREDAEMONRESPONSE']._serialized_end=954
  _globals['_BACKUPRESTOREADMINBACKUPREQUEST']._serialized_start=956
  _globals['_BACKUPRESTOREADMINBACKUPREQUEST']._serialized_end=1020
  _globals['_BACKUPRESTOREADMINBACKUPRESPONSE']._serialized_start=1023
  _globals['_BACKUPRESTOREADMINBACKUPRESPONSE']._serialized_end=1156
  _globals['_BACKUPRESTOREPROFILESNAPSHOTREQUEST']._serialized_start=1159
  _globals['_BACKUPRESTOREPROFILESNAPSHOTREQUEST']._serialized_end=1308
  _globals['_BACKUPRESTOREPROFILESNAPSHOTRESPONSE']._serialized_start=1311
  _globals['_BACKUPRESTOREPROFILESNAPSHOTRESPONSE']._serialized_end=1519
# @@protoc_insertion_point(module_scope)



from typing import Optional

from google.protobuf.json_format import Parse, ParseError

from ..interactive_tree import interactive_tree
from ..tools.cli_tools import *
from ..tools.click_wrappers import WrapperGroup
from ..tools.interactive_actions import ask_question_with_context, invitation_new

#####
# invitation
#####
@interactive_tree.group("invitation", help="send and answer invitations", cls=WrapperGroup)
def invitation_tree():
	pass


#####
# invitation send
#####
@invitation_tree.command("send", help="send a new invitation using an olvid invitation url")
@click.argument("invitation_url", required=True, type=click.STRING)
async def invitation_send(invitation_url: str):
	invitation: datatypes.Invitation = await ClientSingleton.get_client().invitation_new(invitation_url)

	if ClientSingleton.is_script_mode_enabled():
		print(invitation.id)
		return

	prompt: str = "invitation send"
	fg_color: str = "bright_yellow"

	try:
		if not await ask_question_with_context("Do you want to complete invitation process interactively ?", prompt=prompt, fg_color=fg_color):
			print(invitation)
			return

		discussion: Optional[datatypes.Discussion] = await invitation_new(identity_id=ClientSingleton.get_current_identity_id(), invitation=invitation, prompt=prompt, fg_color=fg_color)
		if discussion:
			print(f"You can now send messages to {discussion.title} in discussion {discussion.id}")
		else:
			print("Invitation process finished")
	except click.exceptions.Abort:
		pass


#####
# invitation get
#####
# noinspection PyProtectedMember
@invitation_tree.command("get", help="list current identity invitations")
@click.option("-a", "--all", "get_all", is_flag=True)
@click.argument("invitation_ids", nargs=-1, type=click.INT)
@click.option("-f", "--fields", "fields", type=str)
@click.option("--filter", "filter_", type=str)
async def invitation_get(get_all: bool, invitation_ids: tuple[int], fields: str, filter_: str = ""):
	# build filter
	invitation_filter: datatypes.InvitationFilter = datatypes.InvitationFilter()
	if filter_:
		try:
			parsed_message = Parse(filter_, datatypes.InvitationFilter()._to_native(invitation_filter))
			invitation_filter = datatypes.InvitationFilter._from_native(parsed_message)
		except ParseError as e:
			print_error_message(f"Cannot parse filter: {e}")
			return

	if get_all or not invitation_ids:
		async for invitation in ClientSingleton.get_client().invitation_list(filter=invitation_filter):
			filter_fields_and_print_normal_message(invitation, fields)
	else:
		for invitation_id in invitation_ids:
			invitation = await ClientSingleton.get_client().invitation_get(invitation_id)
			filter_fields_and_print_normal_message(invitation, fields)


#####
# invitation accept
#####
@invitation_tree.command("accept", help="accept an invitation")
@click.argument("invitation_id", required=True, type=click.INT)
async def invitation_accept(invitation_id: int):
	await ClientSingleton.get_client().invitation_accept(invitation_id=invitation_id)
	print_command_result("Invitation accepted")


#####
# invitation decline
#####
@invitation_tree.command("decline", help="decline an invitation")
@click.argument("invitation_id", required=True, type=click.INT)
async def invitation_decline(invitation_id: int):
	await ClientSingleton.get_client().invitation_decline(invitation_id=invitation_id)
	print_command_result("Invitation declined")


#####
# invitation sas
#####
@invitation_tree.command("sas", help="set your future contact sas code")
@click.argument("invitation_id", required=True, type=click.INT)
@click.argument("sas_code", required=True, type=click.STRING)
async def invitation_sas(invitation_id: int, sas_code: str):
	if (len(sas_code) != 4 or not sas_code.isnumeric()):
		raise click.exceptions.BadArgumentUsage("Invalid sas code format")
	await ClientSingleton.get_client().invitation_sas(invitation_id, sas_code)


#####
# invitation rm
#####
@invitation_tree.command("rm", help="delete an invitation")
@click.option("-a", "--all", "delete_all", is_flag=True)
@click.argument("invitation_ids", nargs=-1, type=click.INT)
async def invitation_delete(invitation_ids: tuple[int], delete_all: bool):
	invitation_ids = list(invitation_ids)  # convert tuple to list
	if (not len(invitation_ids) and not delete_all):
		raise click.exceptions.BadArgumentUsage("Specify invitation id")

	if delete_all:
		async for invitation in ClientSingleton.get_client().invitation_list():
			await ClientSingleton.get_client().invitation_delete(invitation_id=invitation.id)
			print_command_result(f"Invitation deleted: {invitation.id}")
	else:
		for invitation_id in invitation_ids:
			await ClientSingleton.get_client().invitation_delete(invitation_id=invitation_id)
			print_command_result(f"Invitation deleted: {invitation_id}")



import logging
import base64
import os

import ressources

from grpc.aio import AioRpcError

from ClientHolder import ClientHolder, ClientWrapper
from utils.tools_group import create_standard_group, create_controlled_group
from olvid import OlvidClient, datatypes, listeners


# noinspection PyProtectedMember
def get_group_seen_by_a_member(original_group: datatypes.Group, member_contact_id: int):
	group: datatypes.Group = original_group._clone()

	# exchange member and owner permissions
	contact_member: datatypes.GroupMember = [m for m in group.members if m.contact_id == member_contact_id][0]
	contact_permissions: datatypes.GroupMemberPermissions = contact_member.permissions
	contact_member.permissions = group.own_permissions
	group.own_permissions = contact_permissions

	# hide ids
	group.id = 0
	for member in group.members:
		member.contact_id = 0
	for pending_member in group.pending_members:
		pending_member.contact_id = 0
		pending_member.pending_member_id = 0

	return group


async def test_disband_group(client_holder: ClientHolder, admin_client: ClientWrapper, group_id: int):
	member_clients: list[ClientWrapper] = [c for c in client_holder.clients if c != admin_client]
	member_contacts: list[datatypes.Contact] = [await admin_client.get_contact_associated_to_another_client(c) for c in
												member_clients]

	original_group: datatypes.Group = await admin_client.group_get(group_id=group_id)

	# create notif-checker bots for group admin
	bots: list[OlvidClient] = [
		admin_client.create_notification_bot(listeners.GroupDeletedListener(handler=admin_client.get_check_content_handler(original_group, notification_type=listeners.NOTIFICATIONS.GROUP_DELETED), count=1)),
		admin_client.create_notification_bot(listeners.DiscussionLockedListener(handler=admin_client.get_check_content_handler(datatypes.Discussion(title=original_group.name), notification_type=listeners.NOTIFICATIONS.DISCUSSION_LOCKED), count=1)),
	]

	# TODO remove this, used to pass a known bug (group type is not stored so we cannot retrieve it when group have been disbanded)
	original_group.type = datatypes.Group.Type.TYPE_UNSPECIFIED

	# create notif-checker bots for members
	for i in range(len(member_clients)):
		member_client = member_clients[i]
		member_contact_id = member_contacts[i].id
		bots.extend([
			member_client.create_notification_bot(listeners.GroupDeletedListener(handler=member_client.get_check_content_handler(get_group_seen_by_a_member(original_group, member_contact_id), notification_type=listeners.NOTIFICATIONS.GROUP_DELETED), count=1)),
			member_client.create_notification_bot(listeners.DiscussionLockedListener(handler=member_client.get_check_content_handler(datatypes.Discussion(title=original_group.name), notification_type=listeners.NOTIFICATIONS.DISCUSSION_LOCKED), count=1)),
		])

	disbanded_group = await admin_client.group_disband(group_id=original_group.id)
	disbanded_group._test_assertion(original_group)

	for bot in bots:
		await bot.wait_for_listeners_end()
		await bot.stop()


async def test_try_do_delete_a_contact_in_group(client_holder: ClientHolder, admin_client: ClientWrapper):
	member_clients: list[ClientWrapper] = [c for c in client_holder.clients if c != admin_client]
	member_contacts: list[datatypes.Contact] = [await admin_client.get_contact_associated_to_another_client(c) for c in member_clients]

	for member_contact in member_contacts:
		try:
			await admin_client.contact_delete(contact_id=member_contact.id)
			assert False, "Was able to delete a contact in a group"
		except AioRpcError:
			continue


async def test_concurrent_group_updates(client_holder: ClientHolder, admin_client: ClientWrapper, group_id: int):
	member_clients: list[ClientWrapper] = [c for c in client_holder.clients if c != admin_client]
	member_contacts: list[datatypes.Contact] = [await admin_client.get_contact_associated_to_another_client(c) for c in member_clients]

	original_group: datatypes.Group = await admin_client.group_get(group_id=group_id)

	group_updated_1: datatypes.Group = original_group._clone()
	group_updated_1.name += "-First Update"
	group_updated_1.description += "-First Update"
	group_updated_2: datatypes.Group = original_group._clone()
	group_updated_2.name = ""
	group_updated_2.description = ""

	# this handler is used to check multiple ordered notifications with different content
	def get_recursive_check_content_handler(client: ClientWrapper, listener_type: type[listeners.GenericNotificationListener], expectations: list[tuple], notification_type: listeners.NOTIFICATIONS):
		def recursive_check_content_handler(*messages):
			if not expectations:
				assert False, "Invalid expectations count"
			handler = client.get_check_content_handler(*(expectations[0]), notification_type=notification_type)
			ret = handler(*messages)

			# create next listener
			if len(expectations) > 1:
				# noinspection PyArgumentList
				client.create_notification_bot(listener_type(handler=get_recursive_check_content_handler(client, listener_type=listener_type, expectations=expectations[1:], notification_type=notification_type), count=1))

			return ret
		return recursive_check_content_handler

	# wait for task 1 and associated notif
	# create notif-checker bots for group admin
	bots: list[OlvidClient] = [
		admin_client.create_notification_bot(
			listeners.GroupNameUpdatedListener(
				handler=get_recursive_check_content_handler(admin_client, listeners.GroupNameUpdatedListener, [(group_updated_1, original_group.name), (group_updated_2, group_updated_1.name)], notification_type=listeners.NOTIFICATIONS.GROUP_NAME_UPDATED),
				count=1)),
		admin_client.create_notification_bot(
			listeners.GroupDescriptionUpdatedListener(
				handler=get_recursive_check_content_handler(admin_client, listeners.GroupDescriptionUpdatedListener, [(group_updated_1, original_group.description), (group_updated_2, group_updated_1.description)], notification_type=listeners.NOTIFICATIONS.GROUP_DESCRIPTION_UPDATED),
				count=1)),
		admin_client.create_notification_bot(
			listeners.DiscussionTitleUpdatedListener(
				handler=get_recursive_check_content_handler(admin_client, listeners.DiscussionTitleUpdatedListener, [(datatypes.Discussion(title=group_updated_1.name), original_group.name), (datatypes.Discussion(title=group_updated_2.name), group_updated_1.name)], notification_type=listeners.NOTIFICATIONS.DISCUSSION_TITLE_UPDATED),
				count=1)),
	]

	# create notif-checker bots for members: members only receive one update notifications (updates are probably too fast to give time to handle intermediary blob)
	for i in range(len(member_clients)):
		member_client = member_clients[i]
		member_contact_id = member_contacts[i].id
		bots.extend([
			member_client.create_notification_bot(
				listeners.GroupNameUpdatedListener(
					handler=member_client.get_check_content_handler(get_group_seen_by_a_member(group_updated_2, member_contact_id), original_group.name, notification_type=listeners.NOTIFICATIONS.GROUP_NAME_UPDATED),
					count=1)),
			member_client.create_notification_bot(
				listeners.GroupDescriptionUpdatedListener(
					handler=member_client.get_check_content_handler(get_group_seen_by_a_member(group_updated_2, member_contact_id), original_group.description, notification_type=listeners.NOTIFICATIONS.GROUP_DESCRIPTION_UPDATED),
					count=1)),
			member_client.create_notification_bot(
				listeners.DiscussionTitleUpdatedListener(
					handler=member_client.get_check_content_handler(datatypes.Discussion(title=group_updated_2.name), original_group.name, notification_type=listeners.NOTIFICATIONS.DISCUSSION_TITLE_UPDATED),
					count=1)),
		])

	# launch concurrent tasks
	task1 = admin_client.group_update(group=group_updated_1)
	task2 = admin_client.group_update(group=group_updated_2)

	group_update_response_1: datatypes.Group = await task1
	group_update_response_1._test_assertion(group_updated_1)

	group_update_response_2: datatypes.Group = await task2
	group_update_response_2._test_assertion(group_updated_2)

	for bot in bots:
		await bot.wait_for_listeners_end()


# noinspection DuplicatedCode
async def test_update_member_permissions(client_holder: ClientHolder, admin_client: ClientWrapper, group_id: int):
	member_clients: list[ClientWrapper] = [c for c in client_holder.clients if c != admin_client]
	member_contacts: list[datatypes.Contact] = [await admin_client.get_contact_associated_to_another_client(c) for c in
												member_clients]

	original_group: datatypes.Group = await admin_client.group_get(group_id=group_id)

	if not original_group.own_permissions.admin:
		return

	# choose a member, save its permissions and update it
	group: datatypes.Group = original_group._clone()
	member_contact_id: int = member_contacts[0].id
	member_index: int = [i for i in range(len(group.members)) if group.members[i].contact_id == member_contact_id][0]
	original_member_permissions: datatypes.GroupMemberPermissions = group.members[member_index].permissions._clone()
	new_member_permissions: datatypes.GroupMemberPermissions = datatypes.GroupMemberPermissions()
	group.members[member_index].permissions = new_member_permissions

	# prepare notification handlers
	bots: list[OlvidClient] = [
		admin_client.create_notification_bot(
			listeners.GroupMemberPermissionsUpdatedListener(
				handler=admin_client.get_check_content_handler(group, group.members[member_index], original_member_permissions, notification_type=listeners.NOTIFICATIONS.GROUP_MEMBER_PERMISSIONS_UPDATED),
				count=1)),
	]

	# update member permissions
	response_group = await admin_client.group_update(group=group)
	response_group._test_assertion(group)

	# wait for notifications
	for bot in bots:
		await bot.wait_for_listeners_end()

	####
	# restore original permissions
	####
	group.members[member_index].permissions = original_member_permissions
	bots: list[OlvidClient] = [
		admin_client.create_notification_bot(
			listeners.GroupMemberPermissionsUpdatedListener(
				handler=admin_client.get_check_content_handler(group, group.members[member_index], new_member_permissions, notification_type=listeners.NOTIFICATIONS.GROUP_MEMBER_PERMISSIONS_UPDATED),
				count=1)),
	]

	# update member permissions
	response_group = await admin_client.group_update(group=group)
	response_group._test_assertion(group)

	# wait for notifications
	for bot in bots:
		await bot.wait_for_listeners_end()


async def test_set_download_unset_photo(client_holder: ClientHolder, admin_client: ClientWrapper, group_id: int):
	admin_group: datatypes.Group = await admin_client.group_get(group_id=group_id)

	#####
	# Set first photo
	#####
	# create notification bots to set photo
	bots: list[OlvidClient] = []
	for client in client_holder.clients:
		group: datatypes.Group = [g async for g in client.group_list(filter=datatypes.GroupFilter(name_search=admin_group.name, description_search=admin_group.description))][-1]
		group_with_photo: datatypes.Group = group._clone()
		group_with_photo.has_a_photo = True
		# we ignore this deprecated field in checks
		group_with_photo.update_in_progress = False
		bots.append(client.create_notification_bot(
			listener=listeners.GroupPhotoUpdatedListener(handler=client.get_check_content_handler(group_with_photo), group_filter=datatypes.GroupFilter(name_search=group.name), count=1)
		))

	# set group photo
	with open("image.png", "wb") as fd:
		fd.write(base64.b64decode(ressources.PNG_IMAGE_AS_B64_2))
	await admin_client.group_set_photo_file(group_id=group_id, file_path="image.png")
	os.remove("image.png")

	for bot in bots:
		await bot.wait_for_listeners_end()
		await bot.stop()

	#####
	# Set another photo
	#####
	# create notification bots to set photo
	bots: list[OlvidClient] = []
	for client in client_holder.clients:
		group: datatypes.Group = [g async for g in client.group_list(filter=datatypes.GroupFilter(name_search=admin_group.name, description_search=admin_group.description))][-1]
		group_with_photo: datatypes.Group = group._clone()
		group_with_photo.has_a_photo = True
		# we ignore this deprecated field in checks
		group_with_photo.update_in_progress = False
		bots.append(client.create_notification_bot(
			listener=listeners.GroupPhotoUpdatedListener(handler=client.get_check_content_handler(group_with_photo), group_filter=datatypes.GroupFilter(name_search=group.name), count=1)
		))

	# set group photo using bytes
	await admin_client.group_set_photo(group_id=group_id, filename="image.png", payload=base64.b64decode(ressources.PNG_IMAGE_AS_B64))

	for bot in bots:
		await bot.wait_for_listeners_end()
		await bot.stop()

	# download photo and check validity
	downloaded_photo: bytes = await admin_client.group_download_photo(group_id=admin_group.id)
	for client in client_holder.clients:
		if client == admin_client:
			continue
		group: datatypes.Group = [g async for g in client.group_list(filter=datatypes.GroupFilter(name_search=admin_group.name, description_search=admin_group.description))][-1]
		other_downloaded_photo: bytes = await client.group_download_photo(group_id=group.id)
		assert len(downloaded_photo) != ressources.PNG_IMAGE_LENGTH_AFTER_DOWNLOAD_DOWNLOADER_SIDE, f"invalid uploader photo size ({len(downloaded_photo)})"
		assert len(other_downloaded_photo) != ressources.PNG_IMAGE_LENGTH_AFTER_DOWNLOAD_UPLOADER_SIDE, f"invalid downloader photo size {len(other_downloaded_photo)}"

	#####
	# Unset photo
	#####
	# create notification bots to unset photo
	bots: list[OlvidClient] = []
	for client in client_holder.clients:
		group: datatypes.Group = [g async for g in client.group_list(filter=datatypes.GroupFilter(name_search=admin_group.name, description_search=admin_group.description))][-1]
		group_without_photo: datatypes.Group = group._clone()
		group_without_photo.has_a_photo = False
		# we ignore this deprecated field in checks
		group_without_photo.update_in_progress = False
		bots.append(client.create_notification_bot(
			listener=listeners.GroupPhotoUpdatedListener(handler=client.get_check_content_handler(group_without_photo), count=1)
		))

	# unset photo
	await admin_client.group_unset_photo(group_id=admin_group.id)

	for bot in bots:
		await bot.wait_for_listeners_end()
		await bot.stop()

	# download photo and check validity
	downloaded_photo: bytes = await admin_client.group_download_photo(group_id=admin_group.id)
	for client in client_holder.clients:
		if client == admin_client:
			continue
		group: datatypes.Group = [g async for g in client.group_list(filter=datatypes.GroupFilter(name_search=admin_group.name, description_search=admin_group.description))][-1]
		other_downloaded_photo: bytes = await client.group_download_photo(group_id=group.id)
		assert len(downloaded_photo) != ressources.PNG_IMAGE_LENGTH_AFTER_DOWNLOAD_DOWNLOADER_SIDE, f"invalid uploader photo size ({len(downloaded_photo)})"
		assert len(other_downloaded_photo) != ressources.PNG_IMAGE_LENGTH_AFTER_DOWNLOAD_UPLOADER_SIDE, f"invalid downloader photo size {len(other_downloaded_photo)}"


async def test_get_identifier(c1: ClientWrapper, group_1: datatypes.Group, client_holder: ClientHolder):
	for c2 in client_holder.clients[1:]:
		group_2: datatypes.Group = [g async for g in c2.group_list(filter=datatypes.GroupFilter(name_search=group_1.name, description_search=group_1.description))][-1]
		identifier_1 = await c1.group_get_bytes_identifier(group_id=group_1.id)
		identifier_2 = await c2.group_get_bytes_identifier(group_id=group_2.id)
		assert identifier_1, "group identifier is empty"
		assert identifier_2, "group identifier is empty"
		assert identifier_1 == identifier_2, "group identifier are not coherent"


async def test_group(client_holder: ClientHolder):
	client = client_holder.clients[0]
	logging.info(f"{client.identity.id}: create normal group")
	group = await create_standard_group(client_holder, client, f"{client.identity.id}-NormalGroupName", group_description=f"{client.identity.id} normal group description")

	await test_get_identifier(client, group, client_holder)

	logging.info(f"{client.identity.id}: delete contact in a group")
	await test_try_do_delete_a_contact_in_group(client_holder, client)
	logging.info(f"{client.identity.id}: concurrent group updates")
	await test_concurrent_group_updates(client_holder, client, group.id)
	logging.info(f"{client.identity.id}: permission updates")
	await test_update_member_permissions(client_holder, client, group.id)
	logging.info(f"{client.identity.id}: set/unset photo")
	await test_set_download_unset_photo(client_holder, client, group.id)

	logging.info(f"{client.identity.id}: disband normal group")
	await test_disband_group(client_holder, client, group.id)

	logging.info(f"{client.identity.id}: create empty name group")
	group = await create_standard_group(client_holder, client, "", "")
	assert group.name == "", "Empty group name is not empty"
	logging.info(f"{client.identity.id}: disband empty name group")
	await test_disband_group(client_holder, client, group.id)

	logging.info(f"{client.identity.id}: create controlled group")
	group = await create_controlled_group(client_holder, client, f"{client.identity.id}-Controlled", group_description="")
	logging.info(f"{client.identity.id}: disband controlled group")
	await test_disband_group(client_holder, client, group.id)


